import '../mixin/selectable.dart';
import '../models/stage_config.dart';
import '../view/stage_items/stage_entity.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:flutter/rendering.dart';
import '../bloc/bloc.dart';
import '../models/stage_session_data.dart';
import '../models/graph_edge.dart';
import '../models/graph_node.dart';
import 'stage_items/wrapper_box.dart';
import 'stage_items/edge.dart';
import '../utils/gesture_transformable.dart';
import 'package:flutter_bloc/flutter_bloc.dart';
import '../utils/node_arrangement.dart';

class Stage extends StatefulWidget {
  final StageConfig config;

  Stage({Key key, this.config}) : super(key: key);

  @override
  _StageState createState() => _StageState();
}

class _StageState extends State<Stage> {
  StageBloc _stageBloc;
  GraphBloc _graphBloc;

  GlobalKey<GestureTransformableState> mainGestureState;

  bool _resetView = true;

  WrapperBox _buildRect(GraphNode node) {
    GlobalKey<WrapperBoxState> key = GlobalKey();
    return WrapperBox(
      key: key,
      node: node,
    );
  }

  Edge _buildEdge(GraphEdge edge) {
    GlobalKey<EdgeState> key = GlobalKey();
    return Edge(
      key: key,
      edgeData: edge,
    );
  }

  @override
  initState() {
    super.initState();
    mainGestureState = GlobalKey();
    _graphBloc = BlocProvider.of<GraphBloc>(context);
    _stageBloc = BlocProvider.of<StageBloc>(context);
  }

  Widget build(BuildContext context) {
    final queryData = MediaQuery.of(context);
    final screenSize = queryData.size;

    final scale = widget.config.initialZoom;

    final Size areaSize = widget.config.stageSize;
    Offset cameraPosition = (widget.config.initialCameraPosition * scale -
            Offset(screenSize.width / 2.0, screenSize.height / 2.0)) *
        -1.0;

    return BlocBuilder<StageBloc, StageSessionData>(
        bloc: _stageBloc,
        builder: (_, StageSessionData state) {
          return Stack(children: <Widget>[
            Positioned(
                left: 0,
                top: 81,
                child: Container(
                    color: _stageBloc
                        .theme.stageOutsideBackgroundColor, // 165,189,214
                    child: GestureTransformable(
                        key: mainGestureState,
                        disableRotation: true,
                        disableTranslation: false,
                        minScale: 0.1,
                        maxScale: 2.5,
                        initialScale: widget.config.initialZoom,
                        reset: _resetView,
                        boundaryRect: Rect.fromLTWH(
                          0,
                          0,
                          areaSize.width,
                          areaSize.height,
                        ),
                        size: screenSize,
                        initialTranslation: cameraPosition,
                        onTapUp: _onStageTapUp,
                        onResetEnd: () {
                          setState(() {
                            _resetView = false;
                          });
                        },
                        child: SizedOverflowBox(
                            alignment: Alignment.topLeft,
                            size: areaSize,
                            child: Container(
                                color:
                                    _stageBloc.theme.stageInsideBackgroundColor,
                                height: areaSize.height,
                                width: areaSize.width,
                                child: BlocBuilder<GraphBloc, GraphState>(
                                    bloc: _graphBloc,
                                    builder: (_, GraphState innerState) {
                                      if (innerState is GraphLoadedState) {
                                        return Stack(
                                            overflow: Overflow.visible,
                                            children: <Widget>[
                                              ...innerState.edges.map(
                                                  (GraphEdge edge) =>
                                                      _buildEdge(edge)),
                                              ...innerState.nodes.map(
                                                  (GraphNode node) =>
                                                      _buildRect(node)),
                                            ]);
                                      }

                                      return Text("Loading...");
                                    })))))),
            Positioned(bottom: 20, left: 20, child: resetButton),
            Positioned(bottom: 20, right: 20, child: deleteButton),
            Positioned(
                top: 0,
                left: 0,
                child: Container(
                    width: screenSize.width,
                    height: 80,
                    child: AppBar(
                      backgroundColor: _stageBloc.theme.navmenuBackgroundColor,
                      title: Row(
                          mainAxisAlignment: MainAxisAlignment.end,
                          children: <Widget>[
                            Row(children: <Widget>[
                              Text('Toggle connection',
                                  textAlign: TextAlign.center,
                                  style: TextStyle(
                                      color: _stageBloc.theme.navmenuTextColor,
                                      fontSize: 14)),
                              Switch(
                                  activeColor: _stageBloc.theme.switchColor,
                                  value: state.connectionEnabled,
                                  onChanged: toggleConnectionMode),
                            ]),
                            Row(children: <Widget>[
                              IconButton(
                                alignment: Alignment.center,
                                icon: const Icon(Icons.save),
                                onPressed: saveStage,
                              )
                            ])
                          ]),
                    )))
          ]);
        });
  }

  FloatingActionButton get deleteButton {
    return FloatingActionButton(
      onPressed: () {
        Selectable<StageEntity> selectedEntity =
            _stageBloc.currentState.selectedEntity;
        if (selectedEntity == null) {
          return;
        }

        if (selectedEntity is WrapperBoxState) {
          _graphBloc.dispatch(RemoveNodeEvent(selectedEntity.node.id));
        } else if (selectedEntity is EdgeState) {
          _graphBloc.dispatch(DisconnectNodesEvent(
              selectedEntity.edgeData.from, selectedEntity.edgeData.to));
        }

        _stageBloc.unselect();
      },
      tooltip: 'Delete an item',
      backgroundColor: _stageBloc.theme.deleteButtonColor,
      child: const Icon(Icons.delete),
    );
  }

  FloatingActionButton get resetButton {
    return FloatingActionButton(
      onPressed: resetView,
      tooltip: 'Reset view',
      backgroundColor: _stageBloc.theme.resetViewButtonColor,
      child: const Icon(Icons.home),
    );
  }

  void saveStage() {
    widget.config.onSave(_graphBloc.graph);
  }

  void resetView() {
    setState(() {
      _resetView = true;
    });
  }

  /*
   * If we click on the link icon, it will switch the linkmode on/off.
   * We can connect boxes only if linkMode equals true.
   * When we are in linkMode, we need to select two boxes by taping them. 
   * After the second selection, the two boxes will be connected.
   */
  void toggleConnectionMode(bool value) {
    _stageBloc.setConnectionMode(value);
  }

  Future<void> showPopup(BuildContext _context, Offset tappedPos) {
    final Matrix4 trans = mainGestureState.currentState.currentTransform;
    final RenderBox overlay = _context.findRenderObject();
    final Size bounds = overlay.size;

    /// transforms to screen, i think it is incorrect...should investigate the computation
    Offset transformed = MatrixUtils.transformPoint(trans, tappedPos);

    final Offset _position = transformed;

    /// only enable popup if the tapped position is inside screen bounds.
    if (false == StageBloc.stageBounds.contains(tappedPos))
      return Future<void>.value(null);

    // smaller rect, the touch area
    // Offset.zero & bounds  // Bigger rect, the entire screen
    return showMenu(

        /// elevation: 100,
        context: _context,
        position: RelativeRect.fromRect(
            Rect.fromLTWH(_position.dx, _position.dy, 40, 40),
            Offset.zero & bounds),
        items: <PopupMenuEntry<int>>[
          PopupMenuItem(
              value: 1,
              child: SizedBox(
                  width: double.infinity,
                  height: 50,
                  child: FlatButton(
                      padding: EdgeInsets.all(0),
                      onPressed: () {
                        GraphNode node = this.widget.config.onAddNew(tappedPos);
                        _graphBloc.dispatch(AddNodeEvent(node: node));

                        /// hide popup menu
                        Navigator.pop<void>(context);
                      },
                      child: Text(
                        "Add New",
                        textAlign: TextAlign.left,
                        style: TextStyle(
                            color: Colors.black, fontWeight: FontWeight.w700),
                      )))),
          PopupMenuItem(
              value: 2,
              child: SizedBox(
                  width: double.infinity,
                  height: 50,
                  child: FlatButton(
                      padding: EdgeInsets.all(0),
                      onPressed: () {
                        Navigator.pop<void>(context);
                        mainGestureState?.currentState
                            ?.moveToLocation(tappedPos);
                      },
                      child: Text(
                        "Zoom",
                        textAlign: TextAlign.left,
                        style: TextStyle(
                            color: Colors.black, fontWeight: FontWeight.w700),
                      )))),
          PopupMenuDivider(height: 5),
          PopupMenuItem(
              value: 3,
              child: SizedBox(
                  width: double.infinity,
                  height: 50,
                  child: FlatButton(
                      padding: EdgeInsets.all(0),
                      onPressed: () {
                        final Offset center =
                            widget.config.stageSize.center(Offset(0, 0));
                        _graphBloc.dispatch(NodeArrangementEvent(
                            center: center,
                            arrangementType: NodeArrange.circle));

                        Navigator.pop<void>(context);
                      },
                      child: Text(
                        "Arrange Circle",
                        textAlign: TextAlign.left,
                        style: TextStyle(
                            color: Colors.black, fontWeight: FontWeight.w700),
                      )))),
          PopupMenuItem(
              value: 3,
              child: SizedBox(
                  width: double.infinity,
                  height: 50,
                  child: FlatButton(
                      padding: EdgeInsets.all(0),
                      onPressed: () {
                        final Offset center =
                            widget.config.stageSize.center(Offset(0, 0));
                        _graphBloc.dispatch(NodeArrangementEvent(
                            center: center,
                            arrangementType: NodeArrange.horizontal));

                        Navigator.pop<void>(context);
                      },
                      child: Text(
                        "Arrange Horizontal",
                        textAlign: TextAlign.left,
                        style: TextStyle(
                            color: Colors.black, fontWeight: FontWeight.w700),
                      )))),
          PopupMenuItem(
              value: 3,
              child: SizedBox(
                  width: double.infinity,
                  height: 50,
                  child: FlatButton(
                      padding: EdgeInsets.all(0),
                      onPressed: () {
                        final Offset center =
                            widget.config.stageSize.center(Offset(0, 0));
                        _graphBloc.dispatch(NodeArrangementEvent(
                            center: center,
                            arrangementType: NodeArrange.vertical));

                        Navigator.pop<void>(context);
                      },
                      child: Text(
                        "Arrange Vertical",
                        textAlign: TextAlign.left,
                        style: TextStyle(
                            color: Colors.black, fontWeight: FontWeight.w700),
                      )))),
          PopupMenuItem(
              value: 3,
              child: SizedBox(
                  width: double.infinity,
                  height: 50,
                  child: FlatButton(
                      padding: EdgeInsets.all(0),
                      onPressed: () {
                        final Offset center =
                            widget.config.stageSize.center(Offset(0, 0));
                        _graphBloc.dispatch(NodeArrangementEvent(
                            center: center,
                            arrangementType: NodeArrange.matrix));

                        Navigator.pop<void>(context);
                      },
                      child: Text(
                        "Arrange Matrix",
                        textAlign: TextAlign.left,
                        style: TextStyle(
                            color: Colors.black, fontWeight: FontWeight.w700),
                      )))),
        ]);
  }

  void _onStageTapUp(TapUpDetails details) {
    _stageBloc.setTappedPos(details.globalPosition);

    if (_stageBloc.currentState.selectedEntity != null) {
      _stageBloc.unselect();
      return;
    }

    showPopup(this.context, details.globalPosition);
  }

  @override
  void dispose() {
    super.dispose();
  }
}
